"""Write and parse W3C datetime.

Each web resource is identified by a URI and may optionally have
other metadata such as timestamp, size, md5. The lastmod property
provides ISO8601 format string access to the timestamp.

The timestamp is assumed to be stored in UTC.
"""

import time
from calendar import timegm
from datetime import datetime
from dateutil import parser as dateutil_parser
import re


def datetime_to_str(dt='now', no_fractions=False):
    """The Last-Modified data in ISO8601 syntax, Z notation.

    The lastmod is stored as unix timestamp which is already
    in UTC. At preesent this code will return 6 decimal digits
    if any fraction of a second is given. It would perhaps be
    better to return only the number of decimal digits necessary,
    up to a resultion of 1 microsecond.

    Special cases:
    - Returns datetime str for now if no parameter given.
    - Returns None if None is supplied.
    """
    if dt is None:
        return None
    elif dt == 'now':
        dt = time.time()
    if no_fractions:
        dt = int(dt)
    else:
        dt += 0.0000001  # improve rounding to microseconds
    return datetime.utcfromtimestamp(dt).isoformat() + 'Z'


def str_to_datetime(s, context='datetime'):
    """Set timestamp from an W3C Datetime Last-Modified value.

    The sitemaps.org specification says that <lastmod> values
    must comply with the W3C Datetime format
    (http://www.w3.org/TR/NOTE-datetime). This is a restricted
    subset of ISO8601. In particular, all forms that include a
    time must include a timezone indication so there is no
    notion of local time (which would be tricky on the web). The
    forms allowed are:

        Year:
          YYYY (eg 1997)
        Year and month:
          YYYY-MM (eg 1997-07)
        Complete date:
          YYYY-MM-DD (eg 1997-07-16)
        Complete date plus hours and minutes:
          YYYY-MM-DDThh:mmTZD (eg 1997-07-16T19:20+01:00)
        Complete date plus hours, minutes and seconds:
          YYYY-MM-DDThh:mm:ssTZD (eg 1997-07-16T19:20:30+01:00)
        Complete date plus hours, minutes, seconds and a decimal fraction
        of a second
          YYYY-MM-DDThh:mm:ss.sTZD (eg 1997-07-16T19:20:30.45+01:00)
        where:
          TZD  = time zone designator (Z or +hh:mm or -hh:mm)

    We do not anticipate the YYYY and YYYY-MM forms being used but
    interpret them as YYYY-01-01 and YYYY-MM-01 respectively. All
    dates are interpreted as having time 00:00:00.0 UTC.

    Datetimes not specified to the level of seconds are intepreted
    as 00.0 seconds.

    *** Extended to make timezone optional ***
    """
    t = None
    if s is None:
        return(t)
    if s == '':
        raise ValueError('Attempt to set empty %s' % (context))
    # Make a date into a full datetime
    m = re.match(r"\d\d\d\d(\-\d\d(\-\d\d)?)?$", s)
    if m is not None:
        if m.group(1) is None:
            s += '-01-01'
        elif m.group(2) is None:
            s += '-01'
        s += 'T00:00:00Z'
    # Now have datetime with timezone info
    m = re.match(r"(.*\d{2}:\d{2}:\d{2})(\.\d+)([^\d].*)?$", s)
    # Chop out fractional seconds if present
    fractional_seconds = 0
    if m is not None:
        s = m.group(1)
        if m.group(3) is not None:
            s += m.group(3)
        fractional_seconds = float(m.group(2))
    # Now check that only allowed formats supplied (the parse
    # function in dateutil is rather lax) and separate out
    # timezone information to be handled separately
    #
    # Seems that one should be able to handle timezone offset
    # with dt.tzinfo module but this has variation in behavior
    # between python 2.6 and 2.7... so do here for now
    m = re.match(r"(\d\d\d\d\-\d\d\-\d\dT\d\d:\d\d(:\d\d)?)(Z|([+-])"
                 r"(\d\d):(\d\d))?$", s)
    if m is None:
        raise ValueError("Bad datetime format (%s)" % s)
    str = m.group(1) + 'Z'
    dt = dateutil_parser.parse(str)
    offset_seconds = 0
    if m.group(3) and m.group(3) != 'Z':
        hh = int(m.group(5))
        mm = int(m.group(6))
        if hh > 23 or mm > 59:
            raise ValueError("Bad timezone offset (%s)" % s)
        offset_seconds = hh * 3600 + mm * 60
        if m.group(4) == '-':
            offset_seconds = -offset_seconds
    # timetuple() ignores timezone information so we have to add in
    # the offset here, and any fractional component of the seconds
    return(timegm(dt.timetuple()) + offset_seconds + fractional_seconds)



# -*- coding: utf-8 -*-
"""Utility functions to support the OCFL Object library."""
import os
import os.path
import re
import sys
try:
    from urllib.parse import quote as urlquote  # py3
except ImportError:                             # pragma: no cover -- py2
    from urllib import quote as urlquote        # pragma: no cover -- py2

from ._version import __version__
from .namaste import find_namastes


NORMALIZATIONS = ['uri', 'md5']  # Must match possibilities in map_filepaths()


def add_object_args(parser):
    """Add Object settings to argparse or argument group instance parser."""
    # Disk scanning
    parser.add_argument('--skip', action='append', default=['README.md', '.DS_Store'],
                        help='directories and files to ignore')
    parser.add_argument('--normalization', '--norm', default=None,
                        help='filepath normalization strategy (None, %s)' %
                        (', '.join(NORMALIZATIONS)))
    # Versioning strategy settings
    parser.add_argument('--no-forward-delta', action='store_true',
                        help='do not use forward deltas')
    parser.add_argument('--no-dedupe', '--no-dedup', action='store_true',
                        help='do not use deduplicate files within a version')
    # Validation settings
    parser.add_argument('--lax-digests', action='store_true',
                        help='allow use of any known digest')
    # Object files
    parser.add_argument('--objdir', '--obj',
                        help='read from or write to OCFL object directory objdir')
    parser.add_argument('--ocfl-version', default='draft',
                        help='OCFL specification version')


def add_shared_args(parser):
    """Add arguments to be shared by any ocfl-py scripts."""
    parser.add_argument('--verbose', '-v', action='store_true',
                        help="be more verbose")
    parser.add_argument('--version', action='store_true',
                        help='Show version number and exit')


def check_shared_args(args):
    """Check arguments set with add_shared_args."""
    if args.version:
        print("%s is part of ocfl-py version %s" % (os.path.basename(sys.argv[0]), __version__))
        sys.exit(0)


def next_version(version):
    """Next version identifier following existing pattern.

    Must deal with both zero-prefixed and non-zero prefixed versions.
    """
    m = re.match(r'''v((\d)\d*)$''', version)
    if not m:
        raise ObjectException("Bad version '%s'" % version)
    next = int(m.group(1)) + 1
    if m.group(2) == '0':
        # Zero-padded version
        next_version = ('v0%0' + str(len(version) - 2) + 'd') % next
        if len(next_version) != len(version):
            raise ObjectException("Version number overflow for zero-padded version %d to %d" % (version, next_version))
        return next_version
    else:
        # Not zero-padded
        return 'v' + str(next)


def remove_first_directory(path):
    """Remove first directory from input path.

    The return value will not have a trailing parh separator, even if
    the input path does. Will return an empty string if the input path
    has just one path segment.
    """
    # FIXME - how to do this efficiently? Current code does complete
    # split and rejoins, excluding the first directory
    rpath = ''
    while True:
        (head, tail) = os.path.split(path)
        if head == path or tail == path:
            break
        else:
            path = head
            rpath = tail if rpath == '' else os.path.join(tail, rpath)
    return rpath


def make_unused_filepath(filepath, used, separator='__'):
    """Find filepath with string appended that makes it disjoint from those in used."""
    n = 1
    while True:
        n += 1
        f = filepath + separator + str(n)
        if f not in used:
            return f


def find_path_type(path):
    """Return a string indicating the type of thing, object or root, at path.

    Looks only at "0=*" Namaste files to determing the path type. Will return
    an error description if not an `object` or storage `root`.
    """
    if not os.path.isdir(path):
        return("does not exist or is not a directory")
    namastes = find_namastes(0, path)
    if len(namastes) == 0:
        return("no 0= declaration file")
    elif len(namastes) > 1:
        return("more than one 0= declaration file")
    m = re.match(r'''ocfl(_object)?_(\d+\.\d+)$''', namastes[0].tvalue)
    if m:
        return('root' if m.group(1) is None else 'object')
    return("unrecognized 0= declaration file 0=%s" % (namastes[0].tvalue))



"""NAMASTE file support.

NAMASTE spec: http://www.cdlib.org/inside/diglib/namaste/namastespec.html

See also command line tool: http://github.com/mjgiarlo/namaste
"""
import os
import os.path
import re


def content_to_tvalue(content):
    """Safe and limited length tvalue from content.

    Return string will be at most 40 characters, trimmed of starting or ending
    whitespace, any input characters not word, dot, hyphen, underscore, colon
    will be converted to underscore.
    """
    content = content.strip()
    return re.sub(r'''[^\w\.\-:]''', '_', content[:40])


def find_namastes(d, dir, max=10):
    """Find NAMASTE files with tag d in dir, return list of Namaste objects.

    max sets a limit on the number of Namaste objects returned, a NamasteException
    will be raised if more than max files with tag d are found.
    """
    prefix = str(d) + '='
    namastes = []
    for filename in os.listdir(dir):
        if filename.startswith(prefix):
            if len(namastes) >= max:
                raise NamasteException("Found too many Namaste files with tag %s in %s" % (str(d), dir))
            namastes.append(Namaste(d, tvalue=filename[len(prefix):]))
    return namastes


def get_namaste(d, dir):
    """Find NAMASTE file with tag d in dir, return Namaste object.

    Raises NamasteException if not exaclty one.
    """
    namastes = find_namastes(d, dir, max=1)
    if len(namastes) != 1:
        raise NamasteException("Failed to find one Namaste file with tag %s in %s" % (str(d), dir))
    return namastes[0]


class NamasteException(Exception):
    """Class for exceptions from Namaste."""

    pass


class Namaste(object):
    """Class implementing NAMASTE specification."""

    def __init__(self, d=0, content='', tvalue=None, tr_func=content_to_tvalue):
        """Initialize Namaste object.

        Parameters:
            d - tag name, D in NAMASTE specification
            content - metadata content of NAMASTE file from which tvalue is derived
            tvalue - explicity set tvalue instead of deriving
            tr_func - function reference used to derive a tvalue from content,
                overriding default content_to_tvalue
        """
        self.d = d
        self.content = content
        self._tvalue = tvalue
        self._tr_func = tr_func

    @property
    def filename(self):
        """Filename of Namaste file."""
        return str(self.d) + '=' + self.tvalue

    @property
    def tvalue(self):
        """tvalue of Namaste file."""
        if self._tvalue is not None:
            return self._tvalue
        else:
            return self._tr_func(self.content)

    def write(self, dir):
        """Write NAMASTE file to dir.

        e.g.
            Namaste(0, 'ocfl_1.0').write(dir)
        """
        with open(os.path.join(dir, self.filename), 'w') as fh:
            fh.write(self.content + "\n")

    def check_content(self, dir):
        """Check that the file content is compatible with the tvalue based on tr_func, else raise NamasteException."""
        filepath = os.path.join(dir, self.filename)
        if self.tvalue == '':
            raise NamasteException("Cannot check Namaste file %s without tvalue being set!" % (filepath))
        if not os.path.isfile(filepath):
            raise NamasteException("Namaste file %s does not exist!" % (filepath))
        with open(filepath, 'r') as fh:
            content = fh.read()
        if self.tvalue != self._tr_func(content):
            raise NamasteException("Content of Namaste file %s doesn't match tvalue %s" % (filepath, self.tvalue))

    def content_ok(self, dir):
        """True is check_content() does not raise an exception."""
        try:
            self.check_content(dir)
        except Exception:
            return False
        return True



"""OCFL Validation Logger.

Handle logging of validation errors and warnings.
"""
import json
import os
import os.path
import re
import logging


class ValidationLogger(object):
    """Class for OCFL ValidationLogger."""

    validation_codes = None

    def __init__(self, show_warnings=False, show_errors=True, lang='en'):
        """Initialize OCFL validation logger."""
        self.show_warnings = show_warnings
        self.show_errors = show_errors
        self.lang = lang
        self.codes = {}
        self.messages = []
        self.num_errors = 0
        self.num_warnings = 0
        self.info = 0
        self.spec = 'https://ocfl.io/draft/spec/'
        if self.validation_codes is None:
            with open(os.path.join(os.path.dirname(__file__), 'data/validation-errors.json'), 'r') as fh:
                self.validation_codes = json.load(fh)

    def error_or_warning(self, code, severity='error', **args):
        """Add error or warning to self.codes."""
        if code in self.validation_codes and 'description' in self.validation_codes[code]:
            desc = self.validation_codes[code]['description']
            lang_desc = None
            if self.lang in desc:
                lang_desc = desc[self.lang]
            elif 'en' in desc:
                lang_desc = desc['en']
            elif len(desc) > 0:
                # first key alphabetically
                lang_desc = desc[sorted(list(desc.keys()))[0]]
            else:
                lang_desc = "Unknown " + severity + ": %s - no description, params (%s)"
            # Add in any parameters
            if 'params' in self.validation_codes[code]:
                params = []
                for param in self.validation_codes[code]['params']:
                    params.append(str(args[param]) if param in args else '???')
                try:
                    lang_desc = lang_desc % tuple(params)
                except TypeError:
                    lang_desc += str(args)
            message = '[' + code + '] ' + lang_desc
        else:
            message = "Unknown " + severity + ": %s - params (%s)" % (code, str(args))
        # Add link to spec
        m = re.match(r'''([EW](\d\d\d))''', code)
        if m and int(m.group(2)) < 200:
            message += ' (see ' + self.spec + '#' + m.group(1) + ')'
        # Store set of codes with last message for that code, and _full_ list of messages
        self.codes[code] = message
        if (severity == 'error' and self.show_errors) or (severity != 'error' and self.show_warnings):
            self.messages.append(message)

    def error(self, code, **args):
        """Add error code to self.codes."""
        self.error_or_warning(code, severity='error', **args)
        self.num_errors += 1

    def warn(self, code, **args):
        """Add warning code to self.codes."""
        self.error_or_warning(code, severity='warning', **args)
        self.num_warnings += 1

    def __str__(self, prefix=''):
        """String of validator status."""
        s = ''
        for message in sorted(self.messages):
            s += prefix + message + '\n'
        return s[:-1]



"""OCFL Inventory Validator.

Code to validate the Python representation of an OCFL Inventory
as read with json.load(). Does not examine anything in storage.
"""
import re

from .digest import digest_regex, normalized_digest
from .validation_logger import ValidationLogger
from .w3c_datetime import str_to_datetime


def get_file_map(inventory, version_dir):
    """Get a map of file in state to files on disk for version_dir in inventory."""
    state = inventory['versions'][version_dir]['state']
    manifest = inventory['manifest']
    file_map = {}
    for digest in state:
        if digest in manifest:
            for file in state[digest]:
                file_map[file] = manifest[digest]
    return file_map


class InventoryValidator(object):
    """Class for OCFL Inventory Validator."""

    def __init__(self, log=None, where='???',
                 lax_digests=False):
        """Initialize OCFL Inventory Validator."""
        self.log = ValidationLogger() if log is None else log
        self.where = where
        # Object state
        self.inventory = None
        self.digest_algorithm = 'sha512'
        self.content_directory = 'content'
        self.all_versions = []
        self.manifest_files = None
        self.head = 'UNKNOWN'
        # Validation control
        self.lax_digests = lax_digests

    def error(self, code, **args):
        """Error with added context."""
        self.log.error(code, where=self.where, **args)

    def warn(self, code, **args):
        """Warning with added context."""
        self.log.warn(code, where=self.where, **args)

    def validate(self, inventory):
        """Validate a given inventory."""
        # Basic structure
        self.inventory = inventory
        if 'id' in inventory:
            iid = inventory['id']
            if type(iid) != str or iid == '':
                self.error("E037")
            elif not re.match(r'''(\w+):.+''', iid):
                self.warn("W005", id=iid)
        else:
            self.error("E036a")
        if 'type' not in inventory:
            self.error("E036b")
        elif inventory['type'] != 'https://ocfl.io/1.0/spec/#inventory':
            self.error("E038")
        if 'digestAlgorithm' not in inventory:
            self.error("E036c")
        elif inventory['digestAlgorithm'] == 'sha512':
            pass
        elif self.lax_digests:
            self.digest_algorithm = inventory['digestAlgorithm']
        elif inventory['digestAlgorithm'] == 'sha256':
            self.warn("W004")
            self.digest_algorithm = inventory['digestAlgorithm']
        else:
            self.error("E039", digest_algorithm=inventory['digestAlgorithm'])
        if 'contentDirectory' in inventory:
            # Careful only to set self.content_directory if value is safe
            cd = inventory['contentDirectory']
            if type(cd) != str or '/' in cd or cd in ['.', '..']:
                self.error("E018")
            else:
                self.content_directory = cd
        if 'manifest' not in inventory:
            self.error("E041a")
        else:
            self.manifest_files = self.validate_manifest(inventory['manifest'])
        if 'versions' not in inventory:
            self.error("E041b")
        else:
            self.all_versions = self.validate_version_sequence(inventory['versions'])
            digests_used = self.validate_versions(inventory['versions'], self.all_versions)
        if 'head' in inventory:
            self.head = 'THERE_ARE_NO_VERSIONS' if len(self.all_versions) == 0 else self.all_versions[-1]
            if inventory['head'] != self.head:
                self.error("E040", got=inventory['head'], expected=self.head)
        else:
            self.error("E036d")
        if 'manifest' in inventory and 'versions' in inventory:
            self.check_digests_present_and_used(self.manifest_files, digests_used)

    def validate_manifest(self, manifest):
        """Validate manifest block in inventory.

        Returns manifest_files, a mapping from file to digest for each file in
        the manifest.
        """
        manifest_files = {}
        manifest_digests = set()
        if type(manifest) != dict:
            self.error('E041c')
        else:
            for digest in manifest:
                m = re.match(self.digest_regex(), digest)
                if not m:
                    self.error('E025a', digest=digest, algorithm=self.digest_algorithm)  # wrong form of digest
                elif type(manifest[digest]) != list:
                    self.error('E092', digest=digest)  # must have path list value
                else:
                    for file in manifest[digest]:
                        norm_digest = normalized_digest(digest, self.digest_algorithm)
                        if norm_digest in manifest_digests:
                            # We have already seen this in different un-normalized form!
                            self.error("E922", digest=norm_digest)
                        else:
                            manifest_files[file] = norm_digest
                            manifest_digests.add(norm_digest)
                            if not self.is_valid_content_path(file):
                                self.error("E042", path=file)
        return manifest_files

    def validate_version_sequence(self, versions):
        """Validate sequence of version names in versions block in inventory.

        Returns an array of in-sequence version directories that are part
        of a valid sequences. May exclude other version directory names that are
        not part of the valid sequence if an error is thrown.
        """
        all_versions = []
        if type(versions) != dict:
            self.error("E044")
            return all_versions
        elif len(versions) == 0:
            self.error("E008b")
            return all_versions
        # Validate version sequence
        # https://ocfl.io/draft/spec/#version-directories
        zero_padded = None
        max_version_num = 999999  # Excessive limit
        if 'v1' in versions:
            fmt = 'v%d'
            zero_padded = False
            all_versions.append('v1')
        else:  # Find padding size
            for n in range(2, 11):
                fmt = 'v%0' + str(n) + 'd'
                vkey = fmt % 1
                if vkey in versions:
                    all_versions.append(vkey)
                    zero_padded = n
                    max_version_num = (10 ** (n - 1)) - 1
                    break
            if not zero_padded:
                self.error("E008a")
                return all_versions
        if zero_padded:
            self.warn("W001")
        # Have v1 and know format, work through to check sequence
        for n in range(2, max_version_num + 1):
            v = (fmt % n)
            if v in versions:
                all_versions.append(v)
            else:
                if len(versions) != (n - 1):
                    self.error("E009")  # Extra version dirs outside sequence
                return all_versions
        # We have now included all possible versions up to the zero padding
        # size, if there are more versions than this number then we must
        # have extra that violate the zero-padding rule or are out of
        # sequence
        if len(versions) > max_version_num:
            self.error("E009")
        return all_versions

    def validate_versions(self, versions, all_versions):
        """Validate versions blocks in inventory.

        Requires as input two things which are assumed to be structurally correct
        from prior basic validation:

          * versions - which is the JSON object (dict) from the inventory
          * all_versions - an ordered list of the versions to look at in versions
                           (all other keys in versions will be ignored)

        Returns a list of digests_used which can then be checked against the
        manifest.
        """
        digests_used = []
        for v in all_versions:
            version = versions[v]
            if 'created' not in version:
                self.error('E048', version=v)  # No created
            elif type(versions[v]['created']) != str:
                self.error('E049d', version=v)  # Bad created
            else:
                created = versions[v]['created']
                try:
                    dt = str_to_datetime(created)
                    if not re.search(r'''(Z|[+-]\d\d:\d\d)$''', created):  # FIXME - kludge
                        self.error('E049a', version=v)
                    if not re.search(r'''T\d\d:\d\d:\d\d''', created):  # FIXME - kludge
                        self.error('E049b', version=v)
                except ValueError as e:
                    self.error('E049c', version=v, description=str(e))
            if 'state' in version:
                digests_used += self.validate_state_block(version['state'], version=v)
            else:
                self.error('E410', version=v)
            if 'message' not in version:
                self.warn('W007a', version=v)
            elif type(version['message']) != str:
                self.error('E048b', version=v)
            if 'user' not in version:
                self.warn('W007b', version=v)
            else:
                user = version['user']
                if type(user) != dict:
                    self.error('E054a', version=v)
                else:
                    if 'name' not in user or type(user['name']) != str:
                        self.error('E054b', version=v)
                    if 'address' not in user:
                        self.warn('W008', version=v)
                    elif type(user['address']) != str:
                        self.error('E054c', version=v)
                    elif not re.match(r'''\w{3,6}:''', user['address']):
                        self.warn('W009', version=v)
        return digests_used

    def validate_state_block(self, state, version):
        """Validate state block in a version in an inventory.

        The version is used only for error reporting.

        Returns a list of content digests referenced in the state block.
        """
        digests = []
        logical_paths = set()
        logical_directories = set()
        if type(state) != dict:
            self.error('E050c', version=version)
        else:
            digest_regex = self.digest_regex()
            for digest in state:
                if not re.match(self.digest_regex(), digest):
                    self.error('E050d', version=version, digest=digest)
                elif type(state[digest]) != list:
                    self.error('E050e', version=version, digest=digest)
                else:
                    for path in state[digest]:
                        if not self.check_logical_path(path, logical_paths, logical_directories):
                            self.error('E051', version=version, digest=digest, path=path)
                    norm_digest = normalized_digest(digest, self.digest_algorithm)
                    if norm_digest in digests:
                        # We have already seen this in different un-normalized form!
                        self.error("E923", version=version, digest=norm_digest)
                    else:
                        digests.append(norm_digest)
            # Check for conflicting logical paths
            for path in logical_directories:
                if path in logical_paths:
                    self.error("E095", version=version, path=path)
        return digests

    def check_logical_path(self, path, logical_paths, logical_directories):
        """Check logical path and accumulate paths/directories for E095 check.

        Neither a leading or trailing slash is allowed which is caught by the
        split and then test for empty.

        logical_paths and logical_directories are expected to sets.
        """
        valid = True
        elements = path.split('/')
        for element in elements:
            if element in ['.', '..', '']:
                valid = False
                break
        # Accumulate paths and directories
        logical_paths.add(path)
        logical_directories.add('/'.join(elements[0:-1]))
        return valid

    def check_digests_present_and_used(self, manifest_files, digests_used):
        """Check all digests in manifest that are needed are present and used."""
        in_manifest = set(manifest_files.values())
        in_state = set(digests_used)
        not_in_manifest = in_state.difference(in_manifest)
        if len(not_in_manifest) > 0:
            self.error("E050a", digests=", ".join(sorted(not_in_manifest)))
        not_in_state = in_manifest.difference(in_state)
        if len(not_in_state) > 0:
            self.error("E050b", digests=", ".join(sorted(not_in_state)))

    def digest_regex(self):
        """A regex for validating un-normalized digest format."""
        try:
            return digest_regex(self.digest_algorithm)
        except ValueError:
            if not self.lax_digests:
                self.error('E026a', digest=self.digest_algorithm)
        # Match anything
        return r'''^.*$'''

    def is_valid_content_path(self, path):
        """True if path is a valid content path."""
        m = re.match(r'''^v\d+/''' + self.content_directory + r'''/(.*)''', path)
        if not m:
            return False
        for element in m.group(1).split('/'):
            if element in ('', '.', '..'):
                return False
        return True

    def validate_as_prior_version(self, prior):
        """Check that prior is a valid InventoryValidator for a prior version of the current inventory object.

        Both inventories are assumed to have been checked for internal consistency.
        """
        # Must have a subset of versions which also check zero padding format etc.
        if not set(prior.all_versions) < set(self.all_versions):
            self.error('E066a', prior_head=prior.head)
        else:
            # Check references to files but realize that there might be different
            # digest algorithms between versions
            for version_dir in prior.all_versions:
                prior_map = get_file_map(prior.inventory, version_dir)
                self_map = get_file_map(self.inventory, version_dir)
                if prior_map.keys() != self_map.keys():
                    self.error('E066b', version_dir=version_dir, prior_head=prior.head)
                else:
                    # Check them all...
                    for file in prior_map:
                        if prior_map[file] != self_map[file]:
                            self.error('E066c', version_dir=version_dir, prior_head=prior.head, file=file)
            # Check metadata
            prior_version = prior.inventory['versions'][version_dir]
            self_version = self.inventory['versions'][version_dir]
            for key in ('created', 'message', 'user'):
                if prior_version.get(key) != self_version.get(key):
                    self.warn('W011', version_dir=version_dir, prior_head=prior.head, key=key)



"""Support for use of Bagit bags with OCFL.

Two scenarios:

1) A Bagit bag is used as the source for the content of one version of an
OCFL object.

2) A Bagit bag is used as the output destination for the content extracted
from one version of an OCFL object.

This code relies upon the https://github.com/LibraryOfCongress/bagit-python
library code to implement "The BagIt File Packaging Format" specification
https://tools.ietf.org/html/rfc8493 . Unlike OCFL, BagIt does not natively
support versioning but is widely used as a transfer format and is thus well
suited to transferring content that might be used to update an OCLF object
or disseminate a particular version.
"""
import bagit
import os.path


class BaggerError(Exception):
    """Exception class for conditions that should abort with message."""

    pass


def bag_as_source(srcbag, metadata):
    """Validate and read metadata from srcbag as input.

    The notion of a bag being valid includes it being complete, ie. not having
    a fetch.txt to provide URLs for files that are not included in local
    filesystem. We thus don't need to test for that case, bagit.is_valid() is
    enough.

    Parameters:
        srcbag - The directory of the bag on disk
        metadata - A VersionMetadata object that will be updated with metadata
            from the bag

    Returns the srcdir for OCFL object content within the bag as it should be
    expressed in the state block.
    """
    bag = bagit.Bag(srcbag)
    if not bag.is_valid():
        raise BaggerError("Source Bagit bag at %s is not valid" % (srcbag))
    # Local arguments override but otherwise take metadata from bag-info.txt
    if not metadata.id and 'External-Identifier' in bag.info:
        metadata.id = bag.info['External-Identifier']
    if not metadata.created and 'Bagging-Date' in bag.info:
        metadata.created = bag.info['Bagging-Date'] + 'T00:00:00Z'  # FIXME - timezone fudge
    if not metadata.message and 'External-Description' in bag.info:
        metadata.message = bag.info['External-Description']
    if not metadata.name and 'Contact-Name' in bag.info:
        metadata.name = bag.info['Contact-Name']
    if not metadata.address and 'Contact-Email' in bag.info:
        metadata.address = 'mailto:' + bag.info['Contact-Email']
    return os.path.join(srcbag, 'data')


def bag_extracted_version(dst, metadata):
    """Bag the extracted files in dst using metadata from metadata.

    Paramaters:
        dst - Destination location for the bag that initially contains the
            content to be bagged (ie. end up in the data directory)
        metadata - A VersionMetadata object that has metadata to add into
            the bag information
    """
    tags = {}
    if metadata.id:
        tags['External-Identifier'] = metadata.id
    if metadata.message:
        tags['External-Description'] = metadata.message
    if metadata.name:
        tags['Contact-Name'] = metadata.name
    if metadata.address and metadata.address.startswith('mailto:'):
        tags['Contact-Email'] = metadata.address[7:]
    bag = bagit.make_bag(dst, bag_info=tags, checksums=['sha512'])



"""Handle UUID URNs encoding with an unencapsualted quadtree path.

Follows similar pattern to Archivematica.
"""
import os
import os.path
import re

from .dispositor import Dispositor


class UUIDQuadtree(Dispositor):
    """Class to support UUID URIs with quadtree directory names."""

    def __init__(self, prefix='urn:uuid:'):
        """Initialize Dispositor."""
        super(UUIDQuadtree, self).__init__()
        self.prefix = prefix

    def encode(self, identifier):
        """NOOP encode identifier."""
        return identifier

    def decode(self, identifier):
        """NOOP decode identifier."""
        return identifier

    def identifier_to_path(self, identifier):
        """Convert identifier to path relative to root.

        Must match prefix:6ba7b810-9dad-11d1-80b4-00c04fd430c8
        """
        path = ''
        if identifier.startswith(self.prefix):
            identifier = identifier[len(self.prefix):]
        else:
            raise Exception("UUIDQuadtree identifier %s does not start with prefix %s" % (identifier, self.prefix))
        match = re.match(r'''([\da-f]{4})([\da-f]{4})\-([\da-f]{4})\-([\da-f]{4})\-([\da-f]{4})\-([\da-f]{4})([\da-f]{4})([\da-f]{4})$''', identifier)
        if not match:
            raise Exception("UUIDQuadtree identifier %s not valid" % (identifer))
        return os.path.join(match.group(1), match.group(2), match.group(3), match.group(4),
                            match.group(5), match.group(6), match.group(7), match.group(8))

    def relative_path_to_identifier(self, path):
        """Convert relative path to identifier."""
        # Combine all directories
        segments = path.split(os.sep)
        if len(segments) != 8:
            raise Exception("Exepected path 8 segments in UUIDQuadtree, got %d from %s" % (len(segments), path))
        for segment in segments:
            if not re.match(r'''[\da-f]{4}$''', segment):
                raise Exception("Bad path segment %s in UUIDQuadtree path %s" % (segment, path))
        return(self.prefix + segments[0] + segments[1] + '-'
               + segments[2] + '-' + segments[3] + '-' + segments[4] + '-'
               + segments[5] + segments[6] + segments[7])



"""Base class for Dispositor objects."""
import os
import os.path
try:
    from urllib.parse import quote_plus, unquote_plus  # py3
except ImportError:                                    # pragma: no cover -- py2
    from urllib import quote_plus, unquote_plus        # pragma: no cover -- py2


class Dispositor(object):
    """Base class for disposition handlers -- let's call them Dispositors."""

    def __init__(self):
        """Initialize Dispositor."""
        pass

    def strip_root(self, path, root):
        """Remove root from path, throw exception on failure."""
        root = root.rstrip(os.sep)  # ditch any trailing path separator
        if os.path.commonprefix((path, root)) == root:
            return os.path.relpath(path, start=root)
        else:
            raise Exception("Path %s is not in root %s" % (path, root))

    def is_valid(self, identifier):
        """True if identifier is valid, always True in this base implementation."""
        return True

    def encode(self, identifier):
        """Encode identifier to get rid of unsafe chars."""
        return quote_plus(identifier)

    def decode(self, identifier):
        """Decode identifier to put back unsafe chars."""
        return unquote_plus(identifier)

    def identifier_to_path(self, identifier):
        """Convert identifier to path relative to some root."""
        raise Exeption("No yet implemented")

    def relative_path_to_identifier(self, path):
        """Convert relative path to identifier."""
        raise Exeption("No yet implemented")

    def path_to_identifier(self, path, root=None):
        """Convert path relative to root to identifier."""
        if root is not None:
            path = self.strip_root(path, root)
        return self.relative_path_to_identifier(path)



"""Metadata for a version of OCFL Object's content."""
import json
import logging
from .w3c_datetime import datetime_to_str


def add_version_metadata_args(parser):
    """Add version metadata settings to argparse instance parser."""
    parser.add_argument('--created', default=None,
                        help='creation time to be used with version(s) added, else '
                             'current time will be recorded')
    parser.add_argument('--message', default=None,
                        help='message to be recorded with version(s) added')
    parser.add_argument('--name', default=None,
                        help='name of user adding version(s) to object')
    parser.add_argument('--address', default=None,
                        help='address of user adding version(s) to object')


class VersionMetadata(object):
    """Class for metadata for a version of OCFL Object's content."""

    def __init__(self, args=None, inventory_file=None, inventory=None, vdir=None):
        """Initialize from command line arguments from argparse."""
        self.id = None
        self.created = None
        self.message = None
        self.name = None
        self.address = None
        if args is not None:
            self.created = args.created
            self.message = args.message
            self.name = args.name
            self.address = args.address
        elif inventory_file is not None:
            self.from_inventory_file(inventory_file, vdir)
        elif inventory is not None:
            self.from_inventory(inventory, vdir)

    def from_inventory_file(self, inventory_file, vdir=None):
        """Initialize from an inventory file."""
        logging.info("Reading metadata for %s from %s" % (vdir, inventory_file))
        with open(inventory_file, 'r') as fh:
            inventory = json.load(fh)
        return self.from_inventory(inventory, vdir, inventory_file)

    def from_inventory(self, inventory, vdir=None, inventory_file=''):
        """Initialize from an inventory object.

        Look for specific version directory vdir if specified, else
        return the head version.

        Parameters:
            inventory - inventory object (dict)
            vdir - explicit version directory, else taken from inventory head
            inventory_file - file name used in error reporting (else '')
        """
        self.id = inventory.get('id', None)
        if 'versions' not in inventory:
            raise Exception("No versions object in inventory %s" % (inventory_file))
        version = None
        if vdir is None:
            if 'head' not in inventory:
                raise Exception("No head version specified in inventory %s" % (inventory_file))
            vdir = inventory['head']
        # Now find version metadata
        if vdir not in inventory['versions']:
            raise Exception("No version block for %s in inventory %s" % (vdir, inventory_file))
        version = inventory['versions'][vdir]
        self.version = vdir
        if 'created' in version:
            self.created = version['created']
        if 'message' in version:
            self.message = version['message']
        if 'user' in version:
            if 'name' in version['user']:
                self.name = version['user']['name']
            if 'address' in version['user']:
                self.address = version['user']['address']

    def as_dict(self, **kwargs):
        """Dictionary object with version metedata."""
        m = {}
        self.add_to_dict(m, **kwargs)
        return m

    def add_to_dict(self, m, **kwargs):
        """Add metadata to dictionary m."""
        m['created'] = self.created if self.created else datetime_to_str()
        if self.message is not None:
            m['message'] = self.message
        if self.name is not None or self.address is not None:
            m['user'] = {'name': self.name}
            if self.address is not None:
                m['user']['address'] = self.address
        # Add any extra values, and they will override instance variables
        for (key, value) in kwargs.items():
            m[key] = value



"""Python implementation of OCFL."""
import sys
from ._version import __version__
from .object import *
from .version import *
from .store import *
from .digest import *
from .disposition import get_dispositor
from .bagger import bag_as_source, bag_extracted_version, BaggerError
from .object_utils import add_object_args, add_shared_args, check_shared_args, find_path_type

if sys.version_info < (3, 6):  # pragma: no cover
    raise Exception("Must use python 3.6 or greater (sorry 2.7, you were a good buddy)!")  # pragma: no cover



"""Handle different storage dispositions."""
from .identity import Identity
from .ntree import Ntree
from .uuid_quadtree import UUIDQuadtree


def get_dispositor(disposition=None):
    """Find Dispositor object for the given disposition."""
    if disposition == 'pairtree':
        return Ntree(n=2)
    elif disposition == 'tripletree':
        return Ntree(n=3)
    elif disposition == 'quadtree':
        return Ntree(n=4)
    elif disposition == 'uuid_quadtree':
        return UUIDQuadtree()
    elif disposition == 'identity':
        return Identity()
    else:
        raise Exception("Unsupported disposition %s, aborting!" % (disposition))



"""Setup for ocfl-py."""
from setuptools import setup, Command
import os
import re

# Extract version number
verfile = open("ocfl/_version.py", "rt").read()
match = re.search(r"^__version__ = '(\d\.\d.\d+(\.\d+)?)'",
                  verfile, re.MULTILINE)
if match:
    version = match.group(1)
else:
    raise RuntimeError("Unable to find version string")


class ShellCommand(Command):
    """Class to with defaults for adding extra shell commnads from setup."""

    user_options = []

    def initialize_options(self):
        """Empty initialize_options."""
        pass

    def finalize_options(self):
        """Empty finalize_options."""
        pass


class Coverage(ShellCommand):
    """Class to allow coverage run from setup."""

    description = "run coverage"

    def run(self):
        """Run coverage program."""
        os.system("coverage run --source=ocfl setup.py test")
        os.system("coverage report")
        os.system("coverage html")
        print("See htmlcov/index.html for details.")


class Readmes(ShellCommand):
    """Class to make readme files for demo runs."""

    description = "make README_ files"

    def run(self):
        """Run coverage program."""
        os.system("python tests/test_ocfl_object_script.py > README_object.md")
        os.system("python tests/test_ocfl_store_script.py > README_store.md")
        print("Built README_object.md and README_store.md.")


setup(
    name='ocfl-py',
    version=version,
    author='Simeon Warner',
    author_email='simeon.warner@cornell.edu',
    packages=['ocfl'],
    package_data={'ocfl': ['data/*']},
    scripts=['ocfl-object.py', 'ocfl-store.py', 'ocfl-validate.py'],
    classifiers=["Development Status :: 2 - Pre-Alpha",
                 "Intended Audience :: Developers",
                 "Operating System :: OS Independent",
                 "Programming Language :: Python",
                 "Programming Language :: Python :: 3.6",
                 "Programming Language :: Python :: 3.7",
                 "Programming Language :: Python :: 3.8",
                 "Topic :: Internet :: WWW/HTTP",
                 "Topic :: Software Development :: "
                 "Libraries :: Python Modules"],
    url='https://github.com/zimeon/ocfl-py',
    description='ocfl-py - A Python implementation of OCFL',
    long_description=open('README').read(),
    install_requires=[
        'pairtree>=0.8.1',
        'dateutils>=0.6.6',
        'bagit>=1.7.0'
    ],
    test_suite="tests",
    cmdclass={
        'coverage': Coverage,
        'readmes': Readmes,
    },
)



# -*- coding: utf-8 -*-
"""Core of OCFL Object library."""
import copy
import hashlib
import json
import os
import os.path
import re
import logging
from shutil import copyfile
import sys
try:
    from urllib.parse import quote as urlquote  # py3
except ImportError:                             # pragma: no cover -- py2
    from urllib import quote as urlquote        # pragma: no cover -- py2

from .digest import file_digest, normalized_digest
from .inventory_validator import InventoryValidator
from .object_utils import remove_first_directory, make_unused_filepath, next_version
from .namaste import Namaste
from .validator import Validator
from .version import VersionMetadata


class ObjectException(Exception):
    """Exception class for OCFL Object."""

    pass


class Object(object):
    """Class for handling OCFL Object data and operations."""

    def __init__(self, id=None, content_directory='content',
                 digest_algorithm='sha512', filepath_normalization='uri',
                 skips=None, forward_delta=True, dedupe=True, lax_digests=False,
                 ocfl_version='draft', fixity=None, verbose=True):
        """Initialize OCFL builder.

        Parameters:
          forward_delta - set False to turn off foward delta
          dedupe - set False to turn off dedupe within versions
          fixity - list of fixity types to add as fixity section
          verbose - set logging level to INFO rather than WARN
        """
        self.id = id
        self.content_directory = content_directory
        self.digest_algorithm = digest_algorithm
        self.filepath_normalization = filepath_normalization
        self.skips = set() if skips is None else set(skips)
        self.forward_delta = forward_delta
        self.dedupe = dedupe
        self.ocfl_version = ocfl_version
        self.fixity = fixity
        self.lax_digests = lax_digests
        self.src_files = {}
        self.log = logging.getLogger(name="ocfl.object")
        self.log.setLevel(level=logging.INFO if verbose else logging.WARN)

    def parse_version_directory(self, dirname):
        """Get version number from version directory name."""
        m = re.match(r'''v(\d{1,5})$''', dirname)
        if not m:
            raise Exception("Bad version directory name: %s" % (dirname))
        v = int(m.group(1))
        if v == 0:
            raise Exception("Bad version directory name: %s, v0 no allowed" % (dirname))
        return v

    def digest(self, filename):
        """Digest for file filename."""
        return file_digest(filename, self.digest_algorithm)

    def map_filepath(self, filepath, vdir, used):
        """Map source filepath to a  name within object.

        The purpose of the mapping might be normalization, sanitization,
        content distribution, or something else.

        Parameters:
          filepath - the source filepath
          vdir - the current version directory
          used - disctionary used to check whether a given vfilepath has
            been used already

        Returns:
          vfilepath - the version filepath for this content that starts
            with vdir/content_directory/
        """
        if self.filepath_normalization == 'uri':
            filepath = urlquote(filepath)
            # also encode any leading period to unhide files
            if filepath[0] == '.':
                filepath = '%2E' + filepath[1:]
        elif self.filepath_normalization == 'md5':
            # Truncated MD5 hash of the _filepath_ as an illustration of diff paths for spec,
            # not sure there could be any real application of this
            filepath = hashlib.md5(filepath.encode('utf-8')).hexdigest()[0:16]
        elif self.filepath_normalization is not None:
            raise Exception("Unknown filepath normalization '%s' requested" % (self.filepath_normalization))
        vfilepath = os.path.join(vdir, self.content_directory, filepath)  # path relative to root, inc v#/content
        # Check we don't already have this vfilepath from many to one normalization,
        # add suffix to distinguish if necessary
        if vfilepath in used:
            vfilepath = make_unused_filepath(vfilepath, used)
        return vfilepath

    def start_inventory(self):
        """Create inventory start with metadata from self."""
        inventory = {
            'id': self.id,
            'type': 'https://ocfl.io/1.0/spec/#inventory',
            'digestAlgorithm': self.digest_algorithm,
            'versions': {},
            'manifest': {}
        }
        # Add contentDirectory if not 'content'
        if self.content_directory != 'content':
            inventory['contentDirectory'] = self.content_directory
        # Add fixity section if requested
        if self.fixity is not None and len(self.fixity) > 0:
            inventory['fixity'] = {}
            for fixity_type in self.fixity:
                inventory['fixity'][fixity_type] = {}
        else:
            self.fixity = None
        return inventory

    def add_version(self, inventory, srcdir, vdir, metadata=None):
        """Add to inventory data for new version based on files in srcdir.

        Parameters:
          inventory - the inventory up to (vdir-1)
          srcdir - the directory path where the files for this new version exist,
             including any version directory that might be present
          vdir - the version directory that these files are being added in
          metadata - a VersionMetadata object

        Returns:
          manifest_to_srcfile - dict mapping from paths in manifest to the full
            path of the source file that should be include in the content for
            this new version
        """
        state = {}  # state for this new version
        manifest = inventory['manifest']
        digests_in_version = {}
        manifest_to_srcfile = {}
        # Special case for testing -- FIXME, should remove this, used only in fixtures/1.0/content/spec-ex-full
        inv_file = srcdir + '_inventory.json'
        if os.path.isfile(inv_file):
            # Read metadata for this version
            metadata.from_inventory_file(inv_file, vdir)
        # Go through all files to find new files in manifest and state for this version
        for (dirpath, dirnames, filenames) in os.walk(srcdir, followlinks=True):
            # Go through filenames, in sort order so the it is deterministic
            # which file is included and which is/are referenced in the case
            # of multiple additions with the same digest
            for filename in sorted(filenames):
                filepath = os.path.join(dirpath, filename)
                sfilepath = os.path.relpath(filepath, srcdir)  # path relative to this version
                vfilepath = self.map_filepath(sfilepath, vdir, manifest_to_srcfile)
                digest = self.digest(filepath)
                # Always add file to state
                if digest not in state:
                    state[digest] = []
                state[digest].append(sfilepath)
                if self.forward_delta and digest in manifest:
                    # We already have this content in a previous version and we are using
                    # forward deltas so do not need to copy in this one
                    pass
                else:
                    # This is a new digest so an addition in this version and
                    # we save the information for later includes
                    if digest not in digests_in_version:
                        digests_in_version[digest] = [vfilepath]
                    elif not self.dedupe:
                        digests_in_version[digest].append(vfilepath)
                    manifest_to_srcfile[vfilepath] = filepath
        # Add any new digests in this version to the manifest
        for digest, paths in digests_in_version.items():
            if digest not in manifest:
                manifest[digest] = paths
            else:
                for p in paths:
                    manifest[digest].append(p)
        # Add extra fixity entries if required
        if self.fixity is not None:
            for fixity_type in self.fixity:
                fixities = inventory['fixity'][fixity_type]
                for digest, vfilepaths in digests_in_version.items():
                    for vfilepath in vfilepaths:
                        fixity_digest = file_digest(manifest_to_srcfile[vfilepath], fixity_type)
                        if fixity_digest not in fixities:
                            fixities[fixity_digest] = [vfilepath]
                        else:
                            fixities[fixity_digest].append(vfilepath)
        # Set head to this latest version, and add this version to inventory
        inventory['head'] = vdir
        inventory['versions'][vdir] = metadata.as_dict(state=state)
        return manifest_to_srcfile

    def build_inventory(self, path, metadata=None):
        """Generator for building an OCFL inventory.

        Yields (vdir, inventory, manifest_to_srcfile) for each version in sequence,
        where vdir is the version directory name, inventory is the inventory for that
        version, and manifest_to_srcfile is a dictionary that maps filepaths in the
        manifest to actual source filepaths.
        """
        inventory = self.start_inventory()
        # Find the versions
        versions = {}
        for vdir in os.listdir(path):
            if vdir in self.skips or not os.path.isdir(os.path.join(path, vdir)):
                continue
            vn = self.parse_version_directory(vdir)
            versions[vn] = vdir
        # Go through versions in order building versions array, deduping if selected
        for vn in sorted(versions.keys()):
            vdir = versions[vn]
            manifest_to_srcfile = self.add_version(inventory, os.path.join(path, vdir), vdir,
                                                   metadata=metadata)
            yield (vdir, inventory, manifest_to_srcfile)

    def write_object_declaration(self, objdir):
        """Write NAMASTE object declaration to objdir."""
        Namaste(0, 'ocfl_object_1.0').write(objdir)

    def write_inventory_and_sidecar(self, objdir, inventory):
        """Write inventory and sidecar to objdir."""
        invfilename = 'inventory.json'
        if not os.path.exists(objdir):
            os.makedirs(objdir)
        invfile = os.path.join(objdir, invfilename)
        with open(invfile, 'w') as fh:
            json.dump(inventory, fh, sort_keys=True, indent=2)
        sidecar = os.path.join(objdir, invfilename + '.' + self.digest_algorithm)
        digest = file_digest(invfile, self.digest_algorithm)
        with open(sidecar, 'w') as fh:
            fh.write(digest + ' ' + invfilename + '\n')

    def build(self, srcdir, metadata=None, objdir=None):
        """Build an OCFL object and write to objdir if set, else print inventories.

        Parameters:
          srcdir - source directory with version sub-directories
          metadata - VersionMetadata object applied to all versions
          objdir - output directory for object (must not already exist), if not
              set then will just write out inventories that would have been
              created
        """
        if self.id is None:
            raise ObjectException("Identifier is not set!")
        if objdir is not None:
            os.makedirs(objdir)
        num_versions = 0
        for (vdir, inventory, manifest_to_srcfile) in self.build_inventory(srcdir, metadata=metadata):
            num_versions += 1
            if objdir is None:
                self.log.warning("### Inventory for %s\n" % (vdir)
                                 + json.dumps(inventory, sort_keys=True, indent=2))
            else:
                self.write_inventory_and_sidecar(os.path.join(objdir, vdir), inventory)
                # Copy files into this version
                for (path, srcfile) in manifest_to_srcfile.items():
                    dstfile = os.path.join(objdir, path)
                    dstpath = os.path.dirname(dstfile)
                    if not os.path.exists(dstpath):
                        os.makedirs(dstpath)
                    copyfile(srcfile, dstfile)
        if objdir is None:
            return
        # Write NAMASTE, inventory and sidecar
        self.write_object_declaration(objdir)
        self.write_inventory_and_sidecar(objdir, inventory)
        logging.info("Built object %s with %s versions" % (self.id, num_versions))

    def create(self, srcdir, metadata=None, objdir=None):
        """Create a new OCFL object with v1 content from srcdir.

        Parameters:
          srcdir - source directory with content for v1
          metadata - VersionMetadata object for v1
          objdir - output directory for object (must not already exist), if not
              set then will just write out inventories that would have been
              created
        """
        if self.id is None:
            raise ObjectException("Identifier is not set!")
        if objdir is not None:
            os.makedirs(objdir)
        inventory = self.start_inventory()
        vdir = 'v1'
        manifest_to_srcfile = self.add_version(inventory, srcdir, vdir,
                                               metadata=metadata)
        if objdir is None:
            self.log.warning("### Inventory for %s\n" % (vdir)
                             + json.dumps(inventory, sort_keys=True, indent=2))
            return
        # Else write out object
        self.write_inventory_and_sidecar(os.path.join(objdir, vdir), inventory)
        # Write NAMASTE, inventory and sidecar
        self.write_object_declaration(objdir)
        self.write_inventory_and_sidecar(objdir, inventory)
        # Write version files
        for digest, paths in inventory['manifest'].items():
            for path in paths:
                srcfile = manifest_to_srcfile[path]
                dstfile = os.path.join(objdir, path)
                dstpath = os.path.dirname(dstfile)
                if not os.path.exists(dstpath):
                    os.makedirs(dstpath)
                copyfile(srcfile, dstfile)
        logging.info("Created OCFL object %s in %s" % (self.id, objdir))

    def update(self, objdir, srcdir=None, metadata=None):
        """Update object creating a new version with content matching srcdir.

        Parameters:
          objdir - directory for object to be update, must contain a valid object!
          srcdir - source directory with version sub-directories
          metadata - VersionMetadata object applied to all versions

        If srcdir is None then the update will be just of metadata and any settings
        (such as using a new digest). There will be no content change between
        versions.
        """
        validator = Validator(check_digests=False, lax_digests=self.lax_digests)
        if not validator.validate(objdir):
            raise ObjectException("Object at '%s' is not valid, aborting" % objdir)
        inventory = self.parse_inventory(objdir)
        self.id = inventory['id']
        old_head = inventory['head']
        versions = inventory['versions']
        head = next_version(old_head)
        logging.info("Will update %s %s -> %s" % (self.id, old_head, head))
        os.mkdir(os.path.join(objdir, head))
        # Is this a request to change the digest algorithm?
        old_digest_algorithm = inventory['digestAlgorithm']
        digest_algorithm = self.digest_algorithm
        if digest_algorithm is None:
            digest_algorithm = old_digest_algorithm
        elif digest_algorithm != old_digest_algorithm:
            logging.info("New version with use %s instead of %s digestAlgorithm" %
                         (digest_algorithm, old_digest_algorithm))
            inventory['digestAlgorithm'] = digest_algorithm
        # Is this a request to change the set of fixity information?
        fixity = self.fixity
        old_fixity = set(inventory['fixity'].keys()) if 'fixity' in inventory else set()
        if fixity is None:
            # Not explicit, carry forward from previous version. Only change will
            # be adding old digest information if we are changing digestAlgorithm
            fixity = old_fixity.copy()
            if digest_algorithm != old_digest_algorithm and old_digest_algorithm not in old_fixity:
                # Add old manifest digests to fixity block
                if 'fixity' not in inventory:
                    inventory['fixity'] = {}
                inventory['fixity'][old_digest_algorithm] = inventory['manifest'].copy()
                fixity.add(old_digest_algorithm)
        else:
            # Fixity to be stored is explicit, may be a change
            fixity = set(fixity)
            if fixity != old_fixity:
                for digest in old_fixity.difference(fixity):
                    inventory['fixity'].pop(digest)
                for digest in fixity.difference(old_fixity):
                    logging.info("FIXME - need to add fixity with digest %s" % digest)
        if fixity != old_fixity:
            logging.info("New version will have %s instead of %s fixity" %
                         (','.join(sorted(fixity)), ','.join(sorted(old_fixity))))
        # Now look at contents, manifest and state
        manifest = copy.deepcopy(inventory['manifest'])
        if digest_algorithm != old_digest_algorithm:
            old_to_new_digest = {}
            new_manifest = {}
            for old_digest, files in manifest.items():
                digest = file_digest(os.path.join(objdir, files[0]), digest_algorithm)
                old_to_new_digest[old_digest] = digest
                for file in files[1:]:
                    # Sanity check that any dupe files also match
                    d = file_digest(os.path.join(objdir, file), digest_algorithm)
                    if d != digest:
                        raise ObjectException("Failed sanity check - files %s and %s should have same %s digest but calculated %s and %s respectively" %
                                              files[0], file, digest_algorithm, digest, d)
                new_manifest[digest] = manifest[old_digest]
            manifest = new_manifest
            # Now update all state blocks
            for vdir in inventory['versions']:
                old_state = inventory['versions'][vdir]['state']
                state = {}
                for old_digest, files in old_state.items():
                    state[old_to_new_digest[old_digest]] = old_state[old_digest]
                inventory['versions'][vdir]['state'] = state
        inventory['manifest'] = manifest
        # Add and remove any contents by comparing srcdir with existing state and manifest
        if srcdir is None:
            # No content Update
            inventory['head'] = head
            state = copy.deepcopy(inventory['versions'][old_head]['state'])
            inventory['versions'][head] = metadata.as_dict(state=state)
        else:
            manifest_to_srcfile = self.add_version(inventory=inventory, srcdir=srcdir, vdir=head, metadata=metadata)
            # Copy files into this version
            for (path, srcfile) in manifest_to_srcfile.items():
                dstfile = os.path.join(objdir, path)
                dstpath = os.path.dirname(dstfile)
                if not os.path.exists(dstpath):
                    os.makedirs(dstpath)
                copyfile(srcfile, dstfile)
        # Write inventory in both root and head version
        self.write_inventory_and_sidecar(os.path.join(objdir, head), inventory)
        self.write_inventory_and_sidecar(objdir, inventory)
        # Delete old root inventory sidecar if we changed digest algorithm
        if digest_algorithm != old_digest_algorithm:
            os.remove(os.path.join(objdir, 'inventory.json.' + old_digest_algorithm))
        logging.info("Updated OCFL object %s in %s by adding %s" % (self.id, objdir, head))

    def _show_indent(self, level, last=False, last_v=False):
        """Indent string for tree view at level for intermediate or last."""
        tree_next = '├── '
        tree_last = '└── '
        tree_pass = '│   '
        tree_indent = '    '
        if level == 0:
            return (tree_last if last else tree_next)
        else:
            return (tree_indent if last else tree_pass) + (tree_last if last_v else tree_next)

    def show(self, objdir):
        """Show OCFL object at objdir."""
        validator = Validator(show_warnings=False,
                              show_errors=True,
                              check_digests=False,
                              lax_digests=self.lax_digests)
        passed = validator.validate(objdir)
        if passed:
            self.log.warning("OCFL object at %s has VALID STRUCTURE (DIGESTS NOT CHECKED) " % (objdir))
        else:
            self.log.warning("OCFL object at %s is INVALID" % (objdir))
        tree = '[' + objdir + ']\n'
        entries = sorted(os.listdir(objdir))
        n = 0
        seen_sidecar = False
        for entry in entries:
            n += 1
            note = entry + ' '
            v_notes = []
            if re.match(r'''v\d+$''', entry):
                seen_v_sidecar = False
                for v_entry in sorted(os.listdir(os.path.join(objdir, entry))):
                    v_note = v_entry + ' '
                    if v_entry == 'inventory.json':
                        pass
                    elif v_entry.startswith('inventory.json.'):
                        if seen_v_sidecar:
                            v_note += '<--- multiple inventory digests?'
                            seen_v_sidecar = True
                    elif v_entry == 'content':
                        num_files = 0
                        for (v_dirpath, v_dirs, v_files) in os.walk(os.path.join(objdir, entry, v_entry), followlinks=True):
                            num_files += len(v_files)
                        v_note += '(%d files)' % num_files
                    else:
                        v_note += '<--- ???'
                    v_notes.append(v_note)
            elif entry in ('0=ocfl_object_1.0', 'inventory.json'):
                pass
            elif entry.startswith('inventory.json.'):
                if seen_sidecar:
                    note += '<--- multiple inventory digests?'
                seen_sidecar = True
            else:
                note += '<--- ???'
            # for (dirpath, dirnames, filenames) in os.walk(, followlinks=True):
            last = (n == len(entries))
            tree += self._show_indent(0, last) + note + "\n"
            nn = 0
            for v_note in v_notes:
                nn += 1
                tree += self._show_indent(1, last, (nn == len(v_notes))) + v_note + "\n"
            self.log.warning("Object tree\n" + tree)

    def validate(self, objdir, show_warnings=True, show_errors=True, check_digests=True):
        """Validate OCFL object at objdir."""
        validator = Validator(show_warnings=show_warnings,
                              show_errors=show_errors,
                              check_digests=check_digests,
                              lax_digests=self.lax_digests)
        passed = validator.validate(objdir)
        messages = str(validator)
        if messages != '':
            print(messages)
        if passed:
            self.log.info("OCFL object at %s is VALID" % (objdir))
        else:
            self.log.info("OCFL object at %s is INVALID" % (objdir))
        return passed

    def extract(self, objdir, version, dstdir):
        """Extract version from object at objdir into dstdir.

        The dstdir itself must not exist but the parent directory must.

        Returns the version block from the inventory.
        """
        # Read inventory, set up version
        inv = self.parse_inventory(objdir)
        if version == 'head':
            version = inv['head']
            logging.info("Object at %s has head %s" % (objdir, version))
        elif version not in inv['versions']:
            raise ObjectException("Object at %s does not include a version '%s'" % (objdir, version))
        # Sanity check on destination
        if os.path.isdir(dstdir):
            raise ObjectException("Target directory %s already exists, aborting!" % (dstdir))

        (parentdir, dir) = os.path.split(os.path.normpath(dstdir))
        if parentdir != '' and not os.path.exists(parentdir):
            raise ObjectException("Destination parent %s does not exist or is not directory" % (parentdir))
        os.mkdir(dstdir)
        # Now extract...
        manifest = inv['manifest']
        state = inv['versions'][version]['state']
        for (digest, logical_files) in state.items():
            existing_file = manifest[digest][0]  # FIXME - pick "best" (closest version?) not first?
            for logical_file in logical_files:
                # FIXME -- need to abstract access so we can, for example, implement S3->local extraction
                logging.debug("Copying %s -> %s" % (digest, logical_file))
                dstfile = os.path.join(dstdir, logical_file)
                dstpath = os.path.dirname(dstfile)
                try:
                    os.makedirs(dstpath)  # exist_ok parameter only in Python 3.2+
                except OSError as e:
                    if not os.path.isdir(dstpath):
                        raise
                copyfile(os.path.join(objdir, existing_file), dstfile)
        logging.info("Extracted %s into %s" % (version, dstdir))
        return VersionMetadata(inventory=inv, vdir=version)

    def parse_inventory(self, path):
        """Read JSON root inventory file for object at path.

        Will validate the inventory and normalize the digests so that the rest
        of the Object methods can assume correctness and matching string digests
        between state and manifest blocks.
        """
        inv_file = os.path.join(path, 'inventory.json')
        with open(inv_file) as fh:
            inventory = json.load(fh)
        # Validate
        iv = InventoryValidator()
        iv.validate(inventory)
        if iv.log.num_errors > 0:
            raise ObjectException("Root inventory is not valid (%d errors)" % iv.log.num_errors)
        digest_algorithm = iv.digest_algorithm
        # Normalize digests in place
        manifest = inventory['manifest']
        for digest in manifest:
            norm_digest = normalized_digest(digest, digest_algorithm)
            if digest != norm_digest:
                manifest[norm_digest] = manifest[digest]
                manifest.remove(digest)
        for v in inventory['versions']:
            state = inventory['versions'][v]['state']
            for digest in state:
                norm_digest = normalized_digest(digest, digest_algorithm)
                if digest != norm_digest:
                    state[norm_digest] = state[digest]
                    state.remove(digest)
        return inventory

    def id_from_inventory(self, path, failure_value='UNKNOWN-ID'):
        """Read JSON root inventory file for object at path and extract id.

        Returns the id from the inventory or failure_value is none can
        be extracted.
        """
        try:
            inventory = self.parse_inventory(path)
            return inventory['id']
        except ObjectException:
            return failure_value
